"""Holds config utilities"""
import os
import re
import subprocess
from typing import Dict, List

import hcl
import jsons
import yaml
from ssm_cache import SSMParameterGroup

from terrawrap.models.wrapper_config import WrapperConfig, AbstractEnvVarConfig, SSMEnvVarConfig
from terrawrap.utils.collection_utils import update

GIT_REPO_REGEX = r"URL.*/([\w-]*)(?:\.git)?"
DEFAULT_REGION = 'us-west-2'
SSM_ENVVAR_CACHE = SSMParameterGroup(max_age=600)


def find_variable_files(path: str) -> List[str]:
    """
    Convenience function for finding all Terraform variable files by walking a given path.
    :param path: The path to the Terraform configuration directory.
    :return: A list of Terraform variable files that can be found by walking down that path, in order of
    discovery from the root of the path.
    """
    variable_files = []

    elements = path.split(os.path.sep)

    cur_path = os.path.sep

    for element in elements:
        cur_path = os.path.join(cur_path, element)
        for file in os.listdir(cur_path):
            if file.endswith(".auto.tfvars"):
                variable_files.append(os.path.join(cur_path, file))

    return variable_files


def find_wrapper_config_files(path: str) -> List[str]:
    """
    Convenience function for finding all wrapper config files by walking a given path.
    :param path: The path to the Terraform configuration directory.
    :return: A list of wrapper config files that can be found by walking down that path, in order of
    discovery from the root of the path.
    """
    wrapper_config_files = []

    elements = path.split(os.path.sep)

    cur_path = os.path.sep

    for element in elements:
        cur_path = os.path.join(cur_path, element)
        for file in os.listdir(cur_path):
            if file.endswith(".tf_wrapper") or file.endswith(".tf_wrapper.yml"):
                wrapper_config_files.append(os.path.join(cur_path, file))

    return wrapper_config_files


def parse_wrapper_configs(wrapper_config_files: List[str]) -> WrapperConfig:
    """
    Function for parsing the Terraform wrapper config file.
    :param wrapper_config_files: A list of file paths to wrapper config files. Config files later in the list
    override those earlier in the list, and are merged with the default config and earlier files.
    :return: A WrapperConfig object representing the accumulated values of all the wrapper config files
    """
    generated_wrapper_config: Dict = {}

    for wrapper_config_path in wrapper_config_files:
        with open(wrapper_config_path) as wrapper_config_file:
            wrapper_config = yaml.safe_load(wrapper_config_file)
            if wrapper_config and isinstance(wrapper_config, dict):
                generated_wrapper_config = update(generated_wrapper_config, wrapper_config)

    wrapper_config_obj: WrapperConfig = jsons.load(generated_wrapper_config, WrapperConfig, strict=True)
    return wrapper_config_obj


def resolve_envvars(envvar_configs: Dict[str, AbstractEnvVarConfig]) -> Dict[str, str]:
    """
    Resolves the 'envvars' section from the wrapper config to actual environment variables that can be easily
    supplied to a command.
    :param envvar_configs: The 'envvars' dictionary from the wrapper config.
    :return: A dictionary representing the environment variables that were resolved, with the key being the
    name of the environment variable and the value being the value of the environment variable.
    """
    resolved_envvars = {}
    for envvar_name, envvar_config in envvar_configs.items():
        if isinstance(envvar_config, SSMEnvVarConfig):
            resolved_envvars[envvar_name] = SSM_ENVVAR_CACHE.parameter(envvar_config.path).value
    return resolved_envvars


def calc_backend_config(path: str, variables: Dict[str, str], wrapper_config: WrapperConfig) -> List[str]:
    """
    Convenience function for calculating the backend config of the given Terraform directory.
    :param path: The path to the directory containing the Terraform config.
    :param variables: The variables derived from the auto.tfvars files.
    :param wrapper_config:
    :return: A dictionary representing the backend configuration for the Terraform directory.
    """
    terraform_bucket = "{region}--mclass--terraform--{account_short_name}".format(
        region=variables.get('region'),
        account_short_name=variables.get('account_short_name')
    )

    output = subprocess.check_output(["git", "remote", "show", "origin", "-n"], cwd=path).decode("utf-8")
    match = re.search(GIT_REPO_REGEX, output)
    if match:
        repo_name = match.group(1)
    else:
        raise RuntimeError("Could not determine git repo name, are we in a git repo?")

    options = {
        'dynamodb_table': variables.get('terraform_lock_table', 'terraform-locking'),
        'encrypt': 'true',
        'key': '%s/%s.tfstate' % (repo_name, path[path.index("/config") + 1:]),
        'region': variables.get('region'),
        'bucket': variables.get('terraform_state_bucket', terraform_bucket),
        'skip_get_ec2_platforms': 'true',
        'skip_region_validation': 'true',
        'skip_credentials_validation': 'true'
    }

    if wrapper_config.backends and wrapper_config.backends.s3:
        # convert the object into a dict so we can append each field to the backend config dynamically
        s3_vars = vars(wrapper_config.backends.s3)
        s3_vars = {key: value for key, value in s3_vars.items() if value is not None}
        options.update(s3_vars)

    backend_config = ['-reconfigure']
    backend_config.extend(['-backend-config=%s=%s' % (key, value) for key, value in options.items()])

    return backend_config


def parse_variable_files(variable_files: List[str]) -> Dict[str, str]:
    """
    Convenience function for parsing variable files and returning the variables as a dictionary
    :param variable_files: List of file paths to variable files. Variable files overwrite files before them.
    :return: A dictionary representing the contents of those variable files.
    """
    variables: Dict = {}

    for variable_file in variable_files:
        with open(variable_file) as var_file:
            variables.update(hcl.load(var_file))

    return variables



"""Data classes to represent the wrapper config file"""
# TODO: convert these classes to dataclasses once we drop support for Python3.6

from enum import Enum
from typing import Dict

import jsons


# pylint: disable=missing-docstring
class EnvVarSource(Enum):
    SSM = 'ssm'


# pylint: disable=missing-docstring
class AbstractEnvVarConfig:
    def __init__(self, source: EnvVarSource):
        self.source = source


# pylint: disable=missing-docstring
class SSMEnvVarConfig(AbstractEnvVarConfig):
    def __init__(self, path: str):
        super().__init__(EnvVarSource.SSM)
        self.path = path


# pylint: disable=missing-docstring
class S3BackendConfig:
    def __init__(self, bucket: str, region: str, dynamodb_table: str = None, role_arn: str = None):
        self.region = region
        self.bucket = bucket
        self.dynamodb_table = dynamodb_table
        self.role_arn = role_arn


# pylint: disable=missing-docstring
class BackendsConfig:
    # pylint: disable=invalid-name
    def __init__(self, s3: S3BackendConfig = None):
        self.s3 = s3


# pylint: disable=unused-argument
def env_var_deserializer(obj_dict, cls, **kwargs):
    """convert a dict to a subclass of AbstractEnvVarConfig"""
    if obj_dict['source'] == EnvVarSource.SSM.value:
        return SSMEnvVarConfig(obj_dict['path'])

    raise RuntimeError('Invalid Source')


jsons.set_deserializer(env_var_deserializer, AbstractEnvVarConfig)


# pylint: disable=missing-docstring
class WrapperConfig:
    def __init__(
            self,
            configure_backend: bool = True,
            pipeline_check: bool = True,
            envvars: Dict[str, AbstractEnvVarConfig] = None,
            backends: BackendsConfig = None
    ):
        self.configure_backend = configure_backend
        self.pipeline_check = pipeline_check
        self.envvars = envvars or {}
        self.backends = backends



"""Module for containing Pipeline Entries"""
import logging
import os
from typing import List, Tuple

from terrawrap.utils.cli import execute_command
from terrawrap.utils.config import find_wrapper_config_files, parse_wrapper_configs, resolve_envvars
from terrawrap.utils.path import get_absolute_path

logger = logging.getLogger(__name__)


class PipelineEntry:
    """Class representing a Pipeline Entry"""

    def __init__(self, path: str, variables: List[str]):
        """
        :param path: The path to the Terraform configuration files to execute.
        :param variables: Any additional variables to set alongside the Terraform command.
        """
        self.path = get_absolute_path(path=path)
        wrapper_config_files = find_wrapper_config_files(self.path)
        wrapper_config = parse_wrapper_configs(wrapper_config_files)
        self.envvars = resolve_envvars(wrapper_config.envvars)
        self.variables = variables

    def execute(self, operation: str, debug: bool = False) -> Tuple[int, List[str]]:
        """
        Function for executing this pipeline entry.
        :param operation: The Terraform operation to execute. IE: apply, plan
        :param debug: True if Terraform debug info should be printed.
        :return: A tuple of the exit code and output of the command.
        """
        command_env = os.environ.copy()
        command_env.update(self.envvars)

        if debug:
            command_env["TF_LOG"] = "DEBUG"

        # We're using --no-resolve-envvars here because we've already resolved the environment variables in
        # the constructor. We are then passing in those environment variables explicitly in the
        # execute_command call below.
        args = ['tf', '--no-resolve-envvars', self.path, operation] + self.variables

        if operation in ["apply", "destroy"]:
            args += ["-auto-approve"]

        return execute_command(
            args,
            print_output=False,
            capture_stderr=True,
            env=command_env
        )



"""Module for interacting with DynamoDB"""

import boto3


class DynamoDB:
    """Class for operating with DynamoDB"""

    def __init__(self, region, client=None):
        self.client = client or boto3.client('dynamodb', region_name=region)

    def upsert_item(self, table_name, primary_key_name, primary_key_value, attribute_name, attribute_value):
        """
        Insert/update items in a DynamoDB table for a specific primary key value.
        The value inserted/updated is of type string.
        :param table_name: DynamoDB table name
        :param primary_key_name: Table's primary key name
        :param primary_key_value: Table's primary key value
        :param attribute_name: Item attribute to be inserted/updated
        :param attribute_value: Attribute value to be inserted/updated
        :return: response
        """
        key = {primary_key_name: {'S': primary_key_value}}
        expression_attribute_values = {':d': {'S': attribute_value}}
        update_expression = 'SET {} = :d'.format(attribute_name)

        return self.client.update_item(
            TableName=table_name,
            Key=key,
            ExpressionAttributeValues=expression_attribute_values,
            UpdateExpression=update_expression
        )



"""Module for containing CLI convenience functions"""
from __future__ import print_function

import subprocess
import tempfile
import logging
from typing import List

from amplify_aws_utils.resource_helper import Jitter

logger = logging.getLogger(__name__)

MAX_RETRIES = 5
RETRIABLE_ERRORS = ['RequestError: send request failed', 'unexpected EOF', 'Throttling',
                    'timeout while waiting for state', 'ServiceUnavailable: Service Unavailable',
                    'failed to decode query XML error response']


# pylint: disable=keyword-arg-before-vararg, too-many-locals
def execute_command(args, print_output=True, capture_stderr=True, print_command=False,
                    retry=False, timeout=15 * 60, *pargs, **kwargs):
    """
    Convenience function for executing a given command and optionally printing the output.
    :param args: List of arguments to execute.
    :param print_output: True if the output of the command should be printed immediately. Defaults to True.
    :param capture_stderr: True if stderr should be captured. Defaults to True.
    :param print_command: True if the command should be printed before executing. Defaults to False.
    :param timeout: Max amount of time to keep retrying to execute command. Defaults to 15 minutes.
    :param retry: Retry a number of times if network errors. Defaults to False.
    :param pargs: Any additional positional arguments to Popen.
    :param kwargs: Any additional keyword arguments to Popen.
    :return: A tuple of the exit code and output of the command.
    """
    max_tries = MAX_RETRIES if retry else 1
    try_count = 0

    jitter = Jitter()
    time_passed = 0
    exit_code = 0
    stdout = ''
    while try_count < max_tries:
        exit_code, stdout = _execute_command(args, print_output, capture_stderr, print_command,
                                             *pargs, **kwargs)

        try_count += 1

        network_errors = _get_retriable_errors(stdout)
        if network_errors and retry:
            logger.warning('Found network errors while running %s command: %s', args, network_errors)
        else:
            # The command either succeeded or failed with a non network error. don't retry
            break

        if time_passed >= timeout:
            raise TimeoutError('Timed out retrying %s command' % args)

        time_passed = jitter.backoff()

    return exit_code, stdout


def _execute_command(args, print_output, capture_stderr, print_command, *pargs, **kwargs):
    """
    Private function for executing a given command and optionally printing the output.
    :param args: List of arguments to execute.
    :param print_output: True if the output of the command should be printed immediately. Defaults to True.
    :param capture_stderr: True if stderr should be captured. Defaults to True.
    :param print_command: True if the command should be printed before executing. Defaults to False.
    :param pargs: Any additional positional arguments to Popen.
    :param kwargs: Any additional keyword arguments to Popen.
    :return: A tuple of the exit code and output of the command.
    """
    stdout_write, stdout_path = tempfile.mkstemp()
    stdout_read = open(stdout_path, "rb")

    if print_command:
        print("Executing: %s" % " ".join(args))

    process = subprocess.Popen(
        args,
        *pargs,
        stdout=stdout_write,
        stderr=stdout_write if capture_stderr else open('/dev/null', 'w'),
        **kwargs
    )

    while True:
        output = stdout_read.read(1).decode()

        if output == '' and process.poll() is not None:
            break

        if print_output and output:
            print(output, end="", flush=True)

    exit_code = process.poll()

    stdout_read.seek(0)
    stdout = [line.decode() for line in stdout_read.readlines()]

    return exit_code, stdout


def _get_retriable_errors(out: List[str]) -> List[str]:
    """Filter line output for retriable errors"""
    return [
        line for line in out
        if any(error in line for error in RETRIABLE_ERRORS)
    ]
